/***********************************************************
 * This software is part of the ProM package * http://www.processmining.org/ * *
 * Copyright (c) 2003-2008 TU/e Eindhoven * and is licensed under the * LGPL
 * License, Version 1.0 * by Eindhoven University of Technology * Department of
 * Information Systems * http://www.processmining.org * *
 ***********************************************************/

package org.processmining.plugins.petrinet.behavioralanalysis;

import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.LivenessInfoConnection;
import org.processmining.models.connections.petrinets.behavioral.NonLiveTransitionsConnection;
import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.ResetInhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.ResetNet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NetAnalysisInformation;
import org.processmining.models.graphbased.directed.petrinet.analysis.NonLiveTransitionsSet;
import org.processmining.models.graphbased.directed.transitionsystem.ReachabilityGraph;
import org.processmining.models.semantics.petrinet.InhibitorNetSemantics;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.models.semantics.petrinet.ResetInhibitorNetSemantics;
import org.processmining.models.semantics.petrinet.ResetNetSemantics;
import org.processmining.models.semantics.petrinet.impl.PetrinetSemanticsFactory;

/**
 * Class to analyze whether a given net is live based on Murata, Tadao. Petri
 * Nets:Properties, Analysis, and Applications. Proceedings of the IEEE vol. 77,
 * No.4, April 1989 . If a net is a free-choice net, it is live iff every siphon
 * contains a marked traps. If it's not, construct a reachability graph. If it
 * has all transitions and every leaf is the same as root node, the net is live.
 * 
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version Dec 10, 2008
 */
@Plugin(name = "Analyze Liveness", returnLabels = { "Liveness info", "Non-live transitions" }, returnTypes = {
		NetAnalysisInformation.LIVENESS.class, NonLiveTransitionsSet.class }, parameterLabels = { "Net", "Marking",
		"Reachability Graph", "Semantic" }, userAccessible = true, mostSignificantResult = -1)
public class LivenessAnalyzer extends AbstractLivenessAnalyzer {

	private Object[] finalizeResult(PluginContext context, PetrinetGraph net, Marking state) {
		context.addConnection(new LivenessInfoConnection(net, state, semantics, info));
		context.getFutureResult(0).setLabel("Liveness Analysis of " + net.getLabel());
		context.getFutureResult(1).setLabel("Non-Live Transitions of " + net.getLabel());

		if (nonLiveTrans == null) {
			// means that the net may not be live, reachability graph cannot be constructed, and non live transition cannot be determined
			nonLiveTrans = new NonLiveTransitionsSet();
		}

		context.addConnection(new NonLiveTransitionsConnection(net, nonLiveTrans, state, semantics));

		return new Object[] { info, nonLiveTrans };

	}

	/**
	 * Variant of net and marking
	 */
	// variant with Petrinet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, Petrinet net, Marking state)
			throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, PetrinetSemanticsFactory
				.regularPetrinetSemantics(Petrinet.class));

		return finalizeResult(context, net, state);
	}

	// variant with InhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, InhibitorNet net, Marking state)
			throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, PetrinetSemanticsFactory
				.regularInhibitorNetSemantics(InhibitorNet.class));
		return finalizeResult(context, net, state);
	}

	// variant with ResetNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, ResetNet net, Marking state)
			throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, PetrinetSemanticsFactory
				.regularResetNetSemantics(ResetNet.class));
		return finalizeResult(context, net, state);
	}

	// variant with ResetInhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, ResetInhibitorNet net, Marking state)
			throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, PetrinetSemanticsFactory
				.regularResetInhibitorNetSemantics(ResetInhibitorNet.class));
		return finalizeResult(context, net, state);
	}

	/**
	 * Variant of net, marking, and semantics
	 */
	// variant with Petrinet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 3 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, Petrinet net, Marking state,
			PetrinetSemantics semantics) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, semantics);
		return finalizeResult(context, net, state);
	}

	// variant with InhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 3 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, InhibitorNet net, Marking state,
			InhibitorNetSemantics semantics) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, semantics);
		return finalizeResult(context, net, state);
	}

	// variant with ResetNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 3 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, ResetNet net, Marking state,
			ResetNetSemantics semantics) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, semantics);
		return finalizeResult(context, net, state);
	}

	// variant with ResetInhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 3 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, ResetInhibitorNet net, Marking state,
			ResetInhibitorNetSemantics semantics) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, null, semantics);
		return finalizeResult(context, net, state);
	}

	/**
	 * Variant of net, marking, and reachability graph
	 */
	// variant with Petrinet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 2 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, Petrinet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, reachabilityGraph, null);
		return finalizeResult(context, net, state);
	}

	// variant with InhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 2 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, InhibitorNet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, reachabilityGraph, null);
		return finalizeResult(context, net, state);
	}

	// variant with ResetNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 2 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, ResetNet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, reachabilityGraph, null);
		return finalizeResult(context, net, state);
	}

	// variant with ResetInhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Liveness", requiredParameterLabels = { 0, 1, 2 })
	public Object[] analyzeLivenessPetriNet(PluginContext context, ResetInhibitorNet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		analyzeLivenessPetriNetPrivate(context, net, state, reachabilityGraph, null);
		return finalizeResult(context, net, state);
	}

}
