/**
 * 
 */
package org.processmining.plugins.petrinet.behavioralanalysis;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.packages.PackageDescriptor;
import org.processmining.framework.packages.PackageManager;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.NonRelaxedSoundTransitionsConnection;
import org.processmining.models.connections.petrinets.behavioral.RelaxedSoundessInfoConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.analysis.NetAnalysisInformation;
import org.processmining.models.graphbased.directed.petrinet.analysis.NetAnalysisInformation.UnDetBool;
import org.processmining.models.graphbased.directed.petrinet.analysis.NonRelaxedSoundTransitionsSet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.models.semantics.petrinet.impl.PetrinetSemanticsFactory;

/**
 * Relaxed soundness checker using LoLA Assuming that the Petri net has
 * interleaving semantics
 * 
 * @author arya
 * 
 */
@Plugin(name = "Analyze Relaxed Soundness (using LoLA for windows)", returnLabels = { "Relaxed soundness info",
		"Unfireable transitions in any sound sequences" }, returnTypes = { NetAnalysisInformation.RELAXEDSOUND.class,
		NonRelaxedSoundTransitionsSet.class }, parameterLabels = { "Petri net", "Marking", "Semantic" }, userAccessible = true, mostSignificantResult = -1)
public class RelaxedSoundnessAnalyzer {

	private static String EXTRAPLACE = "pX"; 
	
	// variant with Petrinet and marking
	@PluginVariant(variantLabel = "Analyze Relaxed Soundness", requiredParameterLabels = { 0, 1 })
	public Object[] analyzeRelaxedSoundPetriNet(PluginContext context, Petrinet net, Marking state)
			throws ConnectionCannotBeObtained {
		context.getConnectionManager().getFirstConnection(InitialMarkingConnection.class, context, net, state);
		return checkRelaxedSoundnessAssumingConnection(context, net, state, PetrinetSemanticsFactory
				.regularPetrinetSemantics(Petrinet.class));
	}
	
	@PluginVariant(variantLabel = "Analyze Relaxed Soundness", requiredParameterLabels = { 0, 1, 2 })
	public Object[] analyzeRelaxedSoundPetriNet(PluginContext context, Petrinet net, Marking state, PetrinetSemantics semantics)
			throws ConnectionCannotBeObtained {
		context.getConnectionManager().getFirstConnection(InitialMarkingConnection.class, context, net, state);
		return checkRelaxedSoundnessAssumingConnection(context, net, state, semantics);
	}

	private Object[] checkRelaxedSoundnessAssumingConnection(PluginContext context, Petrinet net, Marking m,
			PetrinetSemantics semantics) {
		if (m.size() != 1) {
			context.log("Not relaxed sound as the net is not a workflow with one initially marked place");
			return constructResult(context, UnDetBool.FALSE, new NonRelaxedSoundTransitionsSet(), net, m, semantics,
					null, null);
		}

		PackageManager manager = PackageManager.getInstance();
		PackageDescriptor[] packages;
		try {
			packages = manager.findOrInstallPackages("PNAnalysis");
		} catch (Exception exc) {
			context.log("No PNAnalysis package is found");
			return constructResult(context, UnDetBool.UNDETERMINED, new NonRelaxedSoundTransitionsSet(), net, m,
					semantics, null, null);
		}

		String lolaPath = packages[0].getLocalPackageDirectory().getAbsolutePath() + File.separator + "lib"
				+ File.separator;

		// check os, only support win32 and win64
		String os = System.getProperty("os.name").toLowerCase();
		if (os.indexOf("win") >= 0) {
			if (System.getenv("ProgramFiles(x86)") != null) {
				lolaPath += "lola-win64" + File.separator;
			} else if (System.getProperty("os.arch").indexOf("64") != -1) {
				lolaPath += "lola-win64" + File.separator;
			} else {
				lolaPath += "lola-win32" + File.separator;
			}
		} else {
			// not supported platform
			context.log("Unsupported platform for relaxed soundness checking. Property is then undetermined");
			return constructResult(context, UnDetBool.UNDETERMINED, new NonRelaxedSoundTransitionsSet(), net, m,
					semantics, null, null);
		}

		// map each place to label
		Map<String, PetrinetNode> mapIdToNode = new HashMap<String, PetrinetNode>();
		Map<PetrinetNode, String> mapNodeToId = new HashMap<PetrinetNode, String>();

		// after that, construct lola input file and read the result
		String placeStrTemplate = "PLACE ";
		String limiter = "";
		boolean existEndPlace = false;
		String closingQuery = "";

		int id = 0;
		for (Place p : net.getPlaces()) {
			String idStr = "p" + String.valueOf(id);
			mapIdToNode.put(idStr, p);
			mapNodeToId.put(p, idStr);

			placeStrTemplate = placeStrTemplate + limiter + idStr;
			limiter = ", ";
			
			if ((net.getOutEdges(p) == null)||(net.getOutEdges(p).isEmpty())){
				if (!existEndPlace){
					existEndPlace = true;
					closingQuery += ", " + idStr + ":1";
				} else {
					// invalid workflow net
					context.log("There are more than one places without sucessors. This is not a valid workflow net.");
					return constructResult(context, UnDetBool.FALSE, new NonRelaxedSoundTransitionsSet(), net, m,
							semantics, null, null);
				}
			} else {
				closingQuery += ", " + idStr + ":0";
			}

			id++;
		}
		
		if (!existEndPlace){
			// invalid workflow net
			context.log("There is no place without sucessors. This is not a valid workflow net.");
			return constructResult(context, UnDetBool.FALSE, new NonRelaxedSoundTransitionsSet(), net, m,
					semantics, null, null);
		}

		String markingStrTemplate = "MARKING " + mapNodeToId.get(m.iterator().next()) + ":1;";

		Map<String, String> mapTransitionToStrTemplate = new HashMap<String, String>();
		SortedSet<String> unfiredTransitions = new TreeSet<String>();

		for (Transition t : net.getTransitions()) {
			String idStr = "t" + String.valueOf(id);
			mapIdToNode.put(idStr, t);
			mapNodeToId.put(t, idStr);
			unfiredTransitions.add(idStr);

			String strTransTemplate = "";

			// input edge
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> edges = net.getInEdges(t);
			if ((edges != null) && (!edges.isEmpty())) {
				strTransTemplate = "CONSUME ";
				limiter = "";
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : edges) {
					strTransTemplate = strTransTemplate + limiter + mapNodeToId.get(edge.getSource()) + ":1";
					limiter = ", ";
				}
			}

			// output edge
			edges = net.getOutEdges(t);
			if ((edges != null) && (!edges.isEmpty())) {
				strTransTemplate = strTransTemplate + "; PRODUCE ";
				limiter = "";
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getOutEdges(t)) {
					strTransTemplate = strTransTemplate + limiter + mapNodeToId.get(edge.getTarget()) + ":1";
					limiter = ", ";
				}
			}

			mapTransitionToStrTemplate.put(idStr, strTransTemplate);

			id++;
		}

		Set<String> firedTransitions = new HashSet<String>();

		while (!unfiredTransitions.isEmpty()) {
			if (!constructAndCheckSoundness(context, lolaPath, placeStrTemplate, markingStrTemplate, mapTransitionToStrTemplate,
					closingQuery, unfiredTransitions, firedTransitions)) {
				return constructResult(context, UnDetBool.FALSE, new NonRelaxedSoundTransitionsSet(), net, m, semantics,
						unfiredTransitions, mapIdToNode);
			}
			;
		}

		return constructResult(context, UnDetBool.TRUE, new NonRelaxedSoundTransitionsSet(), net, m, semantics,
				unfiredTransitions, mapIdToNode);
	}

	private boolean constructAndCheckSoundness(PluginContext context, String lolaPath, String placeStrTemplate, String markingStrTemplate,
			Map<String, String> mapTransitionToStrTemplate, String closingQuery, SortedSet<String> unfiredTransitions,
			Set<String> firedTransitions) {
		// use lola to check
		String transToBeChecked = unfiredTransitions.iterator().next();
		
		// create file 
		File lolaFile = generateLolaFile(lolaPath + "net.lola", placeStrTemplate, markingStrTemplate, mapTransitionToStrTemplate, closingQuery, transToBeChecked);

		// ask lola and retrieve result
		List<String> result = askLolaToCheckSoundFiringSequence(context, lolaFile, lolaPath, unfiredTransitions);
		if (result != null){
			unfiredTransitions.removeAll(result);
		} else {
			// there is no path
			unfiredTransitions.clear();
			unfiredTransitions.add(transToBeChecked);
		}
		lolaFile.delete();
		
		return (result != null);
	}

	private List<String> askLolaToCheckSoundFiringSequence(PluginContext context, File lolaFile, String lolaPath,
			SortedSet<String> unfiredTransitions) {
		// call Sara and wait until it is finished
		ProcessBuilder b = new ProcessBuilder(lolaPath + "lola-reachability.exe", lolaFile.getAbsolutePath(), "-P");
		b.directory(new File(lolaPath));

		Process processLOLA = null;
		List<String> solution = null;
		try {
			processLOLA = b.start();
			processLOLA.getOutputStream().close();
			processLOLA.getErrorStream().close();

			if (context.getProgress().isCancelled()) {
				throw new InterruptedException("Execution cancelled by user");
			}

			// read Lola's output
			BufferedReader br = new BufferedReader(new InputStreamReader(processLOLA.getInputStream()));

			// parse Lola's output
			String line = "";
			while ((line = br.readLine()) != null) {
				if (line.indexOf("PATH") >= 0) {
					solution = new LinkedList<String>();
					while (((line = br.readLine()) != null)&&(!line.startsWith("lola"))) {
						solution.add(line);
					}
					break;
				}
			}

			// closing
			br.close();
			processLOLA.getInputStream().close();
			processLOLA.destroy();
			processLOLA = null;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} finally {
			if (processLOLA != null) {
				try {
					processLOLA.getInputStream().close();
					processLOLA.destroy();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			}
		}
		return solution;
	}
	

	private File generateLolaFile(String fileName, String placeStrTemplate, String markingStrTemplate,
			Map<String, String> mapTransitionToStrTemplate, String closingQuery, String transToBeChecked) {
		File res = new File(fileName);
		BufferedWriter bwcf = null;
		try {
			bwcf = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(res)));
			bwcf.write(placeStrTemplate);
			bwcf.write(",");
			bwcf.write(EXTRAPLACE);
			bwcf.write(";");
			bwcf.newLine();
			bwcf.write(markingStrTemplate);
			bwcf.newLine();
			
			for (String trans: mapTransitionToStrTemplate.keySet()){
				bwcf.write("TRANSITION ");
				bwcf.write(trans);
				bwcf.newLine();
				bwcf.write(mapTransitionToStrTemplate.get(trans));
				
				// write for this transition
				if (trans.equals(transToBeChecked)){
					bwcf.write(", ");
					bwcf.write(EXTRAPLACE);
					bwcf.write(":1");
				} 
				bwcf.write(";");
				bwcf.newLine();
			}
			
			// write the question
			bwcf.write("ANALYSE MARKING ");
			bwcf.write(EXTRAPLACE);
			bwcf.write(":1");
			bwcf.write(closingQuery);
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		} finally {
			if (bwcf != null) {
				try {
					bwcf.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		return res;
	}

	private Object[] constructResult(PluginContext context, UnDetBool soundnessResult,
			NonRelaxedSoundTransitionsSet transSet, Petrinet net, Marking m, PetrinetSemantics semantics,
			Set<String> unfiredTransitions, Map<String, PetrinetNode> mapIdToNode) {
		NetAnalysisInformation.RELAXEDSOUND relaxedSoundnessInfo = new NetAnalysisInformation.RELAXEDSOUND();
		relaxedSoundnessInfo.setValue(soundnessResult);

		if (soundnessResult.equals(UnDetBool.FALSE)) {
			SortedSet<Transition> result = new TreeSet<Transition>();
			for (String strTrans : unfiredTransitions) {
				result.add((Transition) mapIdToNode.get(strTrans));
			}
			transSet.add(result);
		}

		// add connection
		RelaxedSoundessInfoConnection connection1 = context.addConnection(new RelaxedSoundessInfoConnection(net, m, semantics, relaxedSoundnessInfo));
		connection1.setLabel("Connection between " + net.getLabel() + " and its relaxed soundness info");
		
		NonRelaxedSoundTransitionsConnection connection2 = context.addConnection(new NonRelaxedSoundTransitionsConnection(net, m, semantics, transSet));
		connection2.setLabel("Connection between " + net.getLabel() + " and its non relaxed sound transitions");

		return new Object[] { relaxedSoundnessInfo, transSet };

	}
}
