/***********************************************************
 * This software is part of the ProM package * http://www.processmining.org/ * *
 * Copyright (c) 2003-2008 TU/e Eindhoven * and is licensed under the * LGPL
 * License, Version 1.0 * by Eindhoven University of Technology * Department of
 * Information Systems * http://www.processmining.org * *
 ***********************************************************/

package org.processmining.plugins.petrinet.structuralanalysis;

import java.util.Collection;
import java.util.HashSet;
import java.util.SortedSet;
import java.util.TreeSet;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.structural.ExtFreeChoiceInfoConnection;
import org.processmining.models.connections.petrinets.structural.FreeChoiceInfoConnection;
import org.processmining.models.connections.petrinets.structural.NonExtendedFreeChoiceClustersConnection;
import org.processmining.models.connections.petrinets.structural.NonFreeChoiceClustersConnection;
import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.ResetInhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.ResetNet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NetAnalysisInformation;
import org.processmining.models.graphbased.directed.petrinet.analysis.NonExtendedFreeChoiceClustersSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NonFreeChoiceClustersSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NetAnalysisInformation.UnDetBool;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

/**
 * This class analyze whether a given net is a free choice-net, extended free
 * choice net, or not both Based on Murata, Tadao, Fellow, Lee. Petri Nets:
 * Properties, Analysis, and Applications.
 * 
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version Dec 6, 2008
 */
@Plugin(name = "Analyze Free Choice property and Extended Free Choice", returnLabels = { "FC info", "EFC info",
		"Non-free-choice clusters", "Non-extended-free-choice clusters" }, returnTypes = {
		NetAnalysisInformation.FREECHOICE.class, NetAnalysisInformation.EXTFREECHOICE.class,
		NonFreeChoiceClustersSet.class, NonExtendedFreeChoiceClustersSet.class }, parameterLabels = { "Net" }, userAccessible = true)
public class FreeChoiceAnalyzer {

	@PluginVariant(variantLabel = "Analyze Free Choice and Extended Free Choice property", requiredParameterLabels = { 0 })
	public Object[] analyzeFCAndEFCProperty(PluginContext context, Petrinet net) throws Exception {
		return analyzeFCAndEFCPropertyPriv(context, net);
	}

	@PluginVariant(variantLabel = "Analyze Free Choice and Extended Free Choice property", requiredParameterLabels = { 0 })
	public Object[] analyzeFCAndEFCProperty(PluginContext context, InhibitorNet net) throws Exception {
		return analyzeFCAndEFCPropertyPriv(context, net);
	}

	@PluginVariant(variantLabel = "Analyze Free Choice and Extended Free Choice property", requiredParameterLabels = { 0 })
	public Object[] analyzeFCAndEFCProperty(PluginContext context, ResetNet net) throws Exception {
		return analyzeFCAndEFCPropertyPriv(context, net);
	}

	@PluginVariant(variantLabel = "Analyze Free Choice and Extended Free Choice property", requiredParameterLabels = { 0 })
	public Object[] analyzeFCAndEFCProperty(PluginContext context, ResetInhibitorNet net) throws Exception {
		return analyzeFCAndEFCPropertyPriv(context, net);
	}

	/**
	 * Main method to analyze Free-Choice and Extended Free-Choice property
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be analyzed
	 * @return An array of objects is returned. The first object is of type
	 *         NetAnalysisInformation and contains information whether the net
	 *         is free choice and/or extended free choice. The second object is
	 *         of type NonFreeChoiceClustersSet and contains the non-free-choice
	 *         clusters in the net. The third object has type
	 *         NonExtendedFreeChoiceClustersSet and contains the
	 *         non-extended-free-choice clusters.
	 * @throws Exception
	 */
	private Object[] analyzeFCAndEFCPropertyPriv(PluginContext context, PetrinetGraph net) throws Exception {
		// call main method
		NonFreeChoiceClustersSet nonFCClusters = getNFCClusters(net);
		NonExtendedFreeChoiceClustersSet nonXFCClusters = getNXFCClusters(net);

		// add connection
		NetAnalysisInformation.FREECHOICE fCRes = new NetAnalysisInformation.FREECHOICE();
		fCRes.setValue(nonFCClusters.isEmpty() ? UnDetBool.TRUE : UnDetBool.FALSE);
		context.addConnection(new FreeChoiceInfoConnection(net, fCRes));
		context.getFutureResult(0).setLabel("Free Choice information of " + net.getLabel());

		NetAnalysisInformation.EXTFREECHOICE extFCRes = new NetAnalysisInformation.EXTFREECHOICE();
		extFCRes.setValue(nonXFCClusters.isEmpty() ? UnDetBool.TRUE : UnDetBool.FALSE);
		context.addConnection(new ExtFreeChoiceInfoConnection(net, extFCRes));
		context.getFutureResult(1).setLabel("Extended Free Choice information of " + net.getLabel());

		context.addConnection(new NonFreeChoiceClustersConnection(net, nonFCClusters));
		context.getFutureResult(2).setLabel("Non Free Choice Clusters of " + net.getLabel());
		context.addConnection(new NonExtendedFreeChoiceClustersConnection(net, nonXFCClusters));
		context.getFutureResult(3).setLabel("Non Extended Free Choice Clusters " + net.getLabel());

		return new Object[] { fCRes, extFCRes, nonFCClusters, nonXFCClusters };
	}

	/**
	 * Main method to determine whether a given net is an extended free choice
	 * or not. Extended petri net is an ordinary petrinet so that (P1* intersect
	 * P2*) != null => p1* = p2* for all p1 and p2.
	 * 
	 * @param net
	 *            net to be analyzed
	 * @return the set of non-extended-free-choice clusters.
	 */
	public static NonExtendedFreeChoiceClustersSet getNXFCClusters(PetrinetGraph net) {
		NonFreeChoiceClustersSet clusters = buildClusters(net);
		NonExtendedFreeChoiceClustersSet nonXFCClusters = new NonExtendedFreeChoiceClustersSet();

		for (SortedSet<PetrinetNode> cluster : clusters) {
			if (!isXFC(cluster, net)) {
				nonXFCClusters.add(cluster);
			}
		}

		return nonXFCClusters;
	}

	/**
	 * Main method to determine whether a given net is free choice or not. Free
	 * choice net is an ordinary net such that every arc from a place is either
	 * a unique outgoing arc or a unique incoming arc to a transition
	 * 
	 * @param net
	 *            net to be analyzed
	 * @return The set of non-free-choice clusters
	 */
	public static NonFreeChoiceClustersSet getNFCClusters(PetrinetGraph net) {
		NonFreeChoiceClustersSet clusters = buildClusters(net);
		NonFreeChoiceClustersSet nonFCClusters = new NonFreeChoiceClustersSet();

		for (SortedSet<PetrinetNode> cluster : clusters) {
			if (!isFC(cluster, net)) {
				nonFCClusters.add(cluster);
			}
		}

		return nonFCClusters;
	}

	/**
	 * Determines whether a given cluster is extended free-choice.
	 * 
	 * @param cluster
	 *            The given cluster
	 * @param net
	 *            The corresponding net
	 * @return Whether the cluster is extended free-choice
	 */
	private static boolean isXFC(Collection<PetrinetNode> cluster, PetrinetGraph net) {
		HashSet<Place> places = new HashSet<Place>();
		HashSet<Transition> transitions = new HashSet<Transition>();
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> edges;
		for (PetrinetNode node : cluster) {
			if (node instanceof Place) {
				places.add((Place) node);
			} else {
				transitions.add((Transition) node);
			}
		}
		/**
		 * Every place should have arcs to every transition.
		 */
		for (Place place : places) {
			edges = net.getOutEdges(place);
			if (edges.size() != transitions.size()) {
				return false;
			}
		}
		/**
		 * The weights of the incoming edges to every transition should be
		 * identical.
		 */
		for (Transition transition : transitions) {
			edges = net.getInEdges(transition);
			int weight = ((Arc) edges.iterator().next()).getWeight();
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : edges) {
				if (((Arc) edge).getWeight() != weight) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Determines whether a given cluster is free-choice.
	 * 
	 * @param cluster
	 *            The given cluster
	 * @param net
	 *            The corresponding net
	 * @return Whether the cluster is free-choice
	 */
	private static boolean isFC(Collection<PetrinetNode> cluster, PetrinetGraph net) {
		int nofPlaces = 0, nofTransitions = 0;
		Place place = null;
		for (PetrinetNode node : cluster) {
			if (node instanceof Place) {
				nofPlaces++;
				place = (Place) node;
			} else {
				nofTransitions++;
			}
		}
		if ((nofPlaces > 1) && (nofTransitions > 1)) {
			return false;
		}
		if (nofPlaces == 1) {
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> edges;
			edges = net.getOutEdges(place);
			if (!edges.isEmpty()) {
				int weight = ((Arc) edges.iterator().next()).getWeight();
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : edges) {
					if (((Arc) edge).getWeight() != weight) {
						return false;
					}
				}
			}
		}
		return true;
	}

	/**
	 * Build the set of all cluster for the given net.
	 * 
	 * @param net
	 *            The given net.
	 * @return The set of all clusters.
	 */
	private static NonFreeChoiceClustersSet buildClusters(PetrinetGraph net) {
		NonFreeChoiceClustersSet clusters = new NonFreeChoiceClustersSet();
		Collection<PetrinetNode> clusteredNodes = new HashSet<PetrinetNode>();
		for (Place place : net.getPlaces()) {
			if (!clusteredNodes.contains(place)) {
				TreeSet<PetrinetNode> cluster = new TreeSet<PetrinetNode>();
				buildCluster(place, cluster, net);
				clusteredNodes.addAll(cluster);
				clusters.add(cluster);
			}
		}
		return clusters;
	}

	/**
	 * Builds the cluster for the given node in the given net.
	 * 
	 * @param node
	 *            The given node.
	 * @param cluster
	 *            The resulting cluster.
	 * @param net
	 *            The given net.
	 */
	private static void buildCluster(PetrinetNode node, Collection<PetrinetNode> cluster, PetrinetGraph net) {
		if (!cluster.contains(node)) {
			cluster.add(node);
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> edges;
			if (node instanceof Place) {
				edges = net.getOutEdges(node);
			} else {
				edges = net.getInEdges(node);
			}
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : edges) {
				PetrinetNode otherNode;
				if (node instanceof Place) {
					otherNode = edge.getTarget();
				} else {
					otherNode = edge.getSource();
				}
				buildCluster(otherNode, cluster, net);
			}
		}
	}
}
